/*! \brief Reads one line of SIP message body */
static char *get_body_by_line(const char *line, const char *name, int nameLen)
{
	if (strncasecmp(line, name, nameLen) == 0 && line[nameLen] == '=')
		return ast_skip_blanks(line + nameLen + 1);

	return "";
}

/*! \brief Lookup 'name' in the SDP starting
 * at the 'start' line. Returns the matching line, and 'start'
 * is updated with the next line number.
 */
static const char *get_sdp_iterate(int *start, struct sip_request *req)
{
	int len = strlen(name);

	while (*start < req->sdp_end) {
		const char *r = get_body_by_line(req->line[(*start)++], name, len);
		if (r[0] != '\0')
			return r;
	}

	return "";
}

/*! \brief Lookup 'name' in the SDP between the 'start' line
 * and until the 'stop' line. Returns the matching line, and 'start'
 * is updated with the next line number.
 */
static char get_sdp_line(int *start, int stop, struct sip_request *req, char *line)
{
	char type = "";
	char *tmp = NULL;

	if(stop > req->sdp_end || stop < req->sdp_start) stop = req->sdp_end;

	while (*start < stop) {
		if(sscanf(req->line[(*start)++], "%c=%255s", &type, tmp) == 2) {
			line = ast_skip_blanks(tmp + 2);
			if (option_debug > 2)
				ast_log(LOG_DEBUG, "Got SDP line: %s\n", line);
			return type;
		}
	}

	return "";
}

/*! \brief Process SIP SDP offer, select formats and activate RTP channels
	If offer is rejected, we will not change any properties of the call
 	Return 0 on success, a negative value on errors.
	Must be called after find_sdp().
*/
static int process_sdp(struct sip_pvt *p, struct sip_request *req)
{
	/* iterators for SDP parsing */
	int start = req->sdp_start;
	int next = start;
	int iterator = start;

	/* temporary */
	char type;
	const char *m = NULL;           /* SDP media offer */
	const char *nextm = NULL;
	const char *line = NULL
	int len = -1;

	int portno = -1;		/*!< RTP Audio port number */
	int vportno = -1;		/*!< RTP Video port number */
	int udptlportno = -1;		/*!< UDPTL Image port number */

	/* Peer capability is the capability in the SDP, non codec is RFC2833 DTMF (101) */	
	int peercapability = 0, peernoncodeccapability = 0;
	int vpeercapability = 0, vpeernoncodeccapability = 0;
	struct sockaddr_in sin;		/*!< media socket address */
	struct sockaddr_in vsin;	/*!< Video socket address */

	const char *codecs;
	struct hostent *hp = NULL;	/*!< RTP Audio host IP */
	struct hostent *vhp = NULL;	/*!< RTP video host IP */
	int codec;

	int sendonly = -1;
	int vsendonly = -1;
	int numberofports;
	struct ast_rtp *newaudiortp, *newvideortp;	/* Buffers for codec handling */
	int newjointcapability;				/* Negotiated capability */
	int newpeercapability;
	int newnoncodeccapability;
	int numberofmediastreams = 0;
	int debug = sip_debug_test_pvt(p);

	int last_rtpmap_codec=0;

	if (!p->rtp) {
		ast_log(LOG_ERROR, "Got SDP but have no RTP session allocated.\n");
		return -1;
	}

	/* Initialize the temporary RTP structures we use to evaluate the offer from the peer */
#ifdef LOW_MEMORY
	newaudiortp = ast_threadstorage_get(&ts_audio_rtp, ast_rtp_alloc_size());
#else
	newaudiortp = alloca(ast_rtp_alloc_size());
#endif
	memset(newaudiortp, 0, ast_rtp_alloc_size());
	ast_rtp_new_init(newaudiortp);
	ast_rtp_pt_clear(newaudiortp);

#ifdef LOW_MEMORY
	newvideortp = ast_threadstorage_get(&ts_video_rtp, ast_rtp_alloc_size());
#else
	newvideortp = alloca(ast_rtp_alloc_size());
#endif
	memset(newvideortp, 0, ast_rtp_alloc_size());
	ast_rtp_new_init(newvideortp);
	ast_rtp_pt_clear(newvideortp);


	/* Update our last rtprx when we receive an SDP, too */
	p->lastrtprx = p->lastrtptx = time(NULL); /* XXX why both ? */

	memset(p->offered_media, 0, sizeof(p->offered_media));

	ast_set_flag(&p->flags[0], SIP_NOVIDEO);



	/* Scan for the first media stream (m=) line to limit scanning of globals */
	nextm = get_sdp_iterate(&next, req, "m");
	if(ast_strlen_zero(nextm)) {
		ast_log(LOG_WARNING, "Insufficient information for SDP (m= not found)\n");
		return -1;
	}

	/* Scan session level SDP parameters (lines before first media stream) */
	type = get_sdp_line(&iterator, next, req, line);
	while(!ast_strlen_zero(type)) {
		if(type == 'a')
			process_sdp_a(line);
		else if(type == 'c') {
			if(process_sdp_c(line, hp))
				vph = hp;
		}

		type = get_sdp_line(&iterator, next, req, line);
	}

	/* Scan media stream (m=) specific parameters loop */
	while(!ast_strlen_zero(nextm)) {
		start = next;
		m = nextm;
		nextm = get_sdp_iterate(&next, req, "m");
		int x;
		int audio = FALSE;
		int video = FALSE;
		int image = FALSE;
		numberofports = 1;
		len = -1;

		/* Search for audio media definition */
		if ((sscanf(m, "audio %30d/%30d RTP/AVP %n", &x, &numberofports, &len) == 2 && len > 0) ||
		    (sscanf(m, "audio %30d RTP/AVP %n", &x, &len) == 1 && len > 0)) {
			/* Found audio stream in this media definition */
			audio = TRUE;
			p->offered_media[SDP_AUDIO].offered = TRUE;
			numberofmediastreams++;
			portno = x;

			/* Scan through the RTP payload types specified in a "m=" line: */
			codecs = m + len;
			ast_copy_string(p->offered_media[SDP_AUDIO].text, codecs, sizeof(p->offered_media[SDP_AUDIO].text));
			for (; !ast_strlen_zero(codecs); codecs = ast_skip_blanks(codecs + len)) {
				if (sscanf(codecs, "%30d%n", &codec, &len) != 1) {
					ast_log(LOG_WARNING, "Error in codec string '%s'\n", codecs);
					return -1;
				}
				if (debug)
					ast_verbose("Found RTP audio format %d\n", codec);
				ast_rtp_set_m_type(newaudiortp, codec);
			}
		/* Search for video media definition */
		} else if ((sscanf(m, "video %30d/%30d RTP/AVP %n", &x, &numberofports, &len) == 2 && len > 0) ||
			   (sscanf(m, "video %30d RTP/AVP %n", &x, &len) == 1 && len >= 0)) {
			/* Found video stream in this media definition */
			video = TRUE;
			ast_clear_flag(&p->flags[0], SIP_NOVIDEO);
			p->offered_media[SDP_VIDEO].offered = TRUE;
			numberofmediastreams++;
			vportno = x;

			/* Scan through the RTP payload types specified in a "m=" line: */
			codecs = m + len;
			ast_copy_string(p->offered_media[SDP_VIDEO].text, codecs, sizeof(p->offered_media[SDP_VIDEO].text));
			for (codecs = m + len; !ast_strlen_zero(codecs); codecs = ast_skip_blanks(codecs + len)) {
				if (sscanf(codecs, "%30d%n", &codec, &len) != 1) {
					ast_log(LOG_WARNING, "Error in codec string '%s'\n", codecs);
					return -1;
				}
				if (debug)
					ast_verbose("Found RTP video format %d\n", codec);
				ast_rtp_set_m_type(newvideortp, codec);
			}
		/* Search for image media definition */
		} else if (p->udptl && ((sscanf(m, "image %30d udptl t38%n", &x, &len) == 1 && len > 0) || 
					(sscanf(m, "image %30d UDPTL t38%n", &x, &len) == 1 && len >= 0))) {
			/* Found image stream in this media definition */
			image = TRUE;
			if (debug)
				ast_verbose("Got T.38 offer in SDP in dialog %s\n", p->callid);
			p->offered_media[SDP_IMAGE].offered = TRUE;
			udptlportno = x;
			numberofmediastreams++;
			
			if (p->owner && p->lastinvite) {
				if(p->t38.state != T38_LOCAL_REINVITE) {
					p->t38.state = T38_PEER_REINVITE; /* T38 Offered in re-invite from remote party */
					if (option_debug > 1)
						ast_log(LOG_DEBUG, "T38 state changed to %d on channel %s\n", p->t38.state, p->owner ? p->owner->name : "<none>" );
				}
			} else {
				p->t38.state = T38_PEER_DIRECT; /* T38 Offered directly from peer in first invite */
				p->t38.direct = 1;
				if (option_debug > 1)
					ast_log(LOG_DEBUG, "T38 state changed to %d on channel %s\n", p->t38.state, p->owner ? p->owner->name : "<none>");
			}
		} else {
			ast_log(LOG_WARNING, "Unsupported SDP media type in offer: %s\n", m);
			continue;
		}

		/* Check for number of ports */
		if (numberofports > 1)
			ast_log(LOG_WARNING, "SDP offered %d ports for media, not supported by Asterisk. Will try anyway...\n", numberofports);


		/* Media stream specific parameters */
		while ((type = get_sdp_line(&iterator, next, req, line)) != '\0') {
			/* Check for Media-description-level-address for audio or video */
			if(type == 'c') {
				if(audio) {
					if(!process_sdp_c(line, hp))
						ast_log(LOG_WARNING, "Unable to parse SDP line: c=%s...\n", line);
				}
				else if (video) {
					if(!process_sdp_c(line, vhp))
						ast_log(LOG_WARNING, "Unable to parse SDP line: c=%s...\n", line);
				}
				else {
					ast_log(LOG_WARNING, "SDP offered RTP host for image media, not supported by Asterisk...\n");
					continue;
				}
			}
			/* Check for Media-description-level-parameters */
			else if (type == 'a') {
				/* Audio specific scanning */
				if(audio) {
					if (process_sdp_a_sendonly(a, &sendonly))
						if (option_debug > 2)
							ast_log(LOG_DEBUG, "Processed SDP audio option a=%s\n", line);
					else if (process_sdp_a_audio(line, p, newaudiortp, &last_rtpmap_codec))
						if (option_debug > 2)
							ast_log(LOG_DEBUG, "Processed SDP audio option a=%s\n", line);
					else
						ast_log(LOG_WARNING, "Unsupported SDP audio a=%s...\n", line);
				}
				/* Video specific scanning */
				else if (video) {
					if (process_sdp_a_sendonly(a, &vsendonly))
						if (option_debug > 2)
							ast_log(LOG_DEBUG, "Processed SDP video option a=%s\n", line);
					else if (process_sdp_a_video(line, p, newvideortp, &last_rtpmap_codec)
						if (option_debug > 2)
							ast_log(LOG_DEBUG, "Processed SDP video option a=%s\n", line);
					else
						ast_log(LOG_WARNING, "Unsupported SDP video a=%s...\n", line);
				}
				/* Image (T.38 FAX) specific scanning */
				else if (image) {
					if (process_sdp_a_image(a, p))
						if (option_debug > 2)
							ast_log(LOG_DEBUG, "Processed SDP image option a=%s\n", line);
					else
						ast_log(LOG_WARNING, "Unsupported SDP image a=%s...\n", line);
				}
			}
		}
	}

	if (udptlportno == -1) {
		p->t38.state = T38_DISABLED;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "T38 state changed to %d on channel %s\n", p->t38.state, p->owner ? p->owner->name : "<none>");
	}

	/* Sanity checks */
	if (portno == -1 && vportno == -1 && udptlportno == -1)
		/* No acceptable offer found in SDP  - we have no ports */
		/* Do not change RTP or VRTP if this is a re-invite */
		return -2;

	if (numberofmediastreams > 2)
		/* We have too many fax, audio and/or video media streams, fail this offer */
		return -3;


	/* RTP addresses and ports for audio and video */
	sin.sin_family = AF_INET;
	vsin.sin_family = AF_INET;
	memcpy(&sin.sin_addr, hp->h_addr, sizeof(sin.sin_addr));
	if (vhp)
		memcpy(&vsin.sin_addr, vhp->h_addr, sizeof(vsin.sin_addr));

	/* Setup UDPTL port number */
	if (p->udptl) {
		if (udptlportno > 0) {
			sin.sin_port = htons(udptlportno);
			if (ast_test_flag(&p->flags[0], SIP_NAT) && ast_test_flag(&p->flags[1], SIP_PAGE2_UDPTL_DESTINATION)) {
				struct sockaddr_in peer;
				ast_rtp_get_peer(p->rtp, &peer);
				if (peer.sin_addr.s_addr) {
					memcpy(&sin.sin_addr, &peer.sin_addr, sizeof(sin.sin_addr));
					if (debug) {
						ast_log(LOG_DEBUG, "Peer T.38 UDPTL is set behind NAT and with destination, destination address now %s\n", ast_inet_ntoa(sin.sin_addr));
					}
				}
			}
			ast_udptl_set_peer(p->udptl, &sin);
			if (debug)
				ast_log(LOG_DEBUG,"Peer T.38 UDPTL is at port %s:%d\n",ast_inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
		} else {
			ast_udptl_stop(p->udptl);
			if (debug)
				ast_log(LOG_DEBUG, "Peer doesn't provide T.38 UDPTL\n");
		}
	}

	/* Setup RTP port number */
	if (p->rtp) {
		if (portno > 0) {
			sin.sin_port = htons(portno);
			ast_rtp_set_peer(p->rtp, &sin);
			if (debug)
				ast_verbose("Peer audio RTP is at port %s:%d\n", ast_inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
		} else {
			if (udptlportno > 0) {
				if (debug)
					ast_verbose("Got T.38 Re-invite without audio. Keeping RTP active during T.38 session. Callid %s\n", p->callid);
			} else {
				ast_rtp_stop(p->rtp);
				if (debug)
					ast_verbose("Peer doesn't provide audio. Callid %s\n", p->callid);
			}
		}
	}
	/* Setup video port number */
	if (vportno != -1)
		vsin.sin_port = htons(vportno);
	
	/* Now gather all of the codecs that we are asked for: */
	ast_rtp_get_current_formats(newaudiortp, &peercapability, &peernoncodeccapability);
	ast_rtp_get_current_formats(newvideortp, &vpeercapability, &vpeernoncodeccapability);

	newjointcapability = p->capability & (peercapability | vpeercapability);
	newpeercapability = (peercapability | vpeercapability);
	newnoncodeccapability = p->noncodeccapability & peernoncodeccapability;
		
		
	if (debug) {
		/* shame on whoever coded this.... */
		char s1[SIPBUFSIZE], s2[SIPBUFSIZE], s3[SIPBUFSIZE], s4[SIPBUFSIZE];

		ast_verbose("Capabilities: us - %s, peer - audio=%s/video=%s, combined - %s\n",
			    ast_getformatname_multiple(s1, SIPBUFSIZE, p->capability),
			    ast_getformatname_multiple(s2, SIPBUFSIZE, newpeercapability),
			    ast_getformatname_multiple(s3, SIPBUFSIZE, vpeercapability),
			    ast_getformatname_multiple(s4, SIPBUFSIZE, newjointcapability));

		ast_verbose("Non-codec capabilities (dtmf): us - %s, peer - %s, combined - %s\n",
			    ast_rtp_lookup_mime_multiple(s1, SIPBUFSIZE, p->noncodeccapability, 0, 0),
			    ast_rtp_lookup_mime_multiple(s2, SIPBUFSIZE, peernoncodeccapability, 0, 0),
			    ast_rtp_lookup_mime_multiple(s3, SIPBUFSIZE, newnoncodeccapability, 0, 0));
	}
	if (!newjointcapability) {
		/* If T.38 was not negotiated either, totally bail out... */
		if (!p->t38.jointcapability || !udptlportno) {
			ast_log(LOG_NOTICE, "No compatible codecs, not accepting this offer!\n");
			/* Do NOT Change current setting */
			return -1;
		} else {
			if (option_debug > 2)
				ast_log(LOG_DEBUG, "Have T.38 but no audio codecs, accepting offer anyway\n");
			return 0;
		}
	}

	/* We are now ready to change the sip session and p->rtp and p->vrtp with the offered codecs, since
		they are acceptable */
	p->jointcapability = newjointcapability;	        /* Our joint codec profile for this call */
	p->peercapability = newpeercapability;		        /* The other sides capability in latest offer */
	p->jointnoncodeccapability = newnoncodeccapability;	/* DTMF capabilities */

	ast_rtp_pt_copy(p->rtp, newaudiortp);
	if (p->vrtp)
		ast_rtp_pt_copy(p->vrtp, newvideortp);

	if (ast_test_flag(&p->flags[0], SIP_DTMF) == SIP_DTMF_AUTO) {
		ast_clear_flag(&p->flags[0], SIP_DTMF);
		if (newnoncodeccapability & AST_RTP_DTMF) {
			/* XXX Would it be reasonable to drop the DSP at this point? XXX */
			ast_set_flag(&p->flags[0], SIP_DTMF_RFC2833);
			/* Since RFC2833 is now negotiated we need to change some properties of the RTP stream */
			ast_rtp_setdtmf(p->rtp, 1);
			ast_rtp_setdtmfcompensate(p->rtp, ast_test_flag(&p->flags[1], SIP_PAGE2_RFC2833_COMPENSATE));
		} else {
			ast_set_flag(&p->flags[0], SIP_DTMF_INBAND);
		}
	}

	/* Setup audio port number */
	if (p->rtp && sin.sin_port) {
		ast_rtp_set_peer(p->rtp, &sin);
		if (debug)
			ast_verbose("Peer audio RTP is at port %s:%d\n", ast_inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
	}

	/* Setup video port number */
	if (p->vrtp && vsin.sin_port) {
		ast_rtp_set_peer(p->vrtp, &vsin);
		if (debug) 
			ast_verbose("Peer video RTP is at port %s:%d\n", ast_inet_ntoa(vsin.sin_addr), ntohs(vsin.sin_port));
	}

	/* Ok, we're going with this offer */
	if (option_debug > 1) {
		char buf[SIPBUFSIZE];
		ast_log(LOG_DEBUG, "We're settling with these formats: %s\n", ast_getformatname_multiple(buf, SIPBUFSIZE, p->jointcapability));
	}

	if (!p->owner) 	/* There's no open channel owning us so we can return here. For a re-invite or so, we proceed */
		return 0;

	if (option_debug > 3)
		ast_log(LOG_DEBUG, "We have an owner, now see if we need to change this call\n");

	if (!(p->owner->nativeformats & p->jointcapability) && (p->jointcapability & AST_FORMAT_AUDIO_MASK)) {
		if (debug) {
			char s1[SIPBUFSIZE], s2[SIPBUFSIZE];
			ast_log(LOG_DEBUG, "Oooh, we need to change our audio formats since our peer supports only %s and not %s\n", 
				ast_getformatname_multiple(s1, SIPBUFSIZE, p->jointcapability),
				ast_getformatname_multiple(s2, SIPBUFSIZE, p->owner->nativeformats));
		}
		p->owner->nativeformats = ast_codec_choose(&p->prefs, p->jointcapability, 1) | (p->capability & vpeercapability);
		ast_set_read_format(p->owner, p->owner->readformat);
		ast_set_write_format(p->owner, p->owner->writeformat);
	}
	
	if (ast_test_flag(&p->flags[1], SIP_PAGE2_CALL_ONHOLD) && sin.sin_addr.s_addr && (!sendonly || sendonly == -1)) {
		ast_queue_control(p->owner, AST_CONTROL_UNHOLD);
		/* Activate a re-invite */
		ast_queue_frame(p->owner, &ast_null_frame);
	} else if (!sin.sin_addr.s_addr || (sendonly && sendonly != -1)) {
		ast_queue_control_data(p->owner, AST_CONTROL_HOLD, 
				       S_OR(p->mohsuggest, NULL),
				       !ast_strlen_zero(p->mohsuggest) ? strlen(p->mohsuggest) + 1 : 0);
		if (sendonly)
			ast_rtp_stop(p->rtp);
		/* RTCP needs to go ahead, even if we're on hold!!! */
		/* Activate a re-invite */
		ast_queue_frame(p->owner, &ast_null_frame);
	}

	/* Manager Hold and Unhold events must be generated, if necessary */
	if (ast_test_flag(&p->flags[1], SIP_PAGE2_CALL_ONHOLD) && sin.sin_addr.s_addr && (!sendonly || sendonly == -1))
		change_hold_state(p, req, FALSE, sendonly);
	else if (!sin.sin_addr.s_addr || (sendonly && sendonly != -1))
		change_hold_state(p, req, TRUE, sendonly);

	return 0;
}







static int process_sdp_c(char *c, hostent* hp)
{
	char host[258];
	ast_hostent* ast_hp = NULL;

	/* Check for Media-description-level-address */
	if (!ast_strlen_zero(c)) {
		if (sscanf(c, "IN IP4 %255s", host) != 1) {
			ast_log(LOG_WARNING, "Invalid secondary host in c= line, '%s'\n", c);
			return FALSE;
		} else {
			if (!(hp = ast_gethostbyname(host, &ast_hp))) {
				ast_log(LOG_WARNING, "Unable to lookup RTP Audio host in secondary c= line, '%s'\n", c);
				return FALSE;
			}
			return TRUE;
		}
	}
	return FALSE;
}

static int process_sdp_a(char *a, struct sip_pvt *p, struct ast_rtp *newaudiortp, struct ast_rtp *newvideortp, int *sendonly, int *last_rtpmap_codec)
{
	int found = FALSE;

	/* Scan unsupported options if debug > 1 */
	if (option_debug > 1 && process_sdp_a_unsupported(a)) {
		found = TRUE;
	}
	else if (process_sdp_a_sendonly(a, &sendonly)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "Processed global SDP option a=%s\n", a);
	}
	else if (process_sdp_a_audio(a, p, newaudiortp, last_rtpmap_codec)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "Processed global SDP audio option a=%s\n", a);
	}
	else if (process_sdp_a_video(a, p, newvideortp, last_rtpmap_codec)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "Processed global SDP video option a=%s\n", a);
	}
	else if (process_sdp_a_image(a, p)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "Processed global SDP image option a=%s\n", a);
	}

	return found;
}

static int process_sdp_a_unsupported(char *a)
{
	int found = FALSE;

	/* If we're debugging, check for unsupported sdp options */
	if (!strncasecmp(a, "rtcp:", (size_t) 5)) {
		if (debug)
			ast_verbose("Got unsupported a:rtcp in SDP offer \n");
		found = TRUE;
	} else if (!strncasecmp(a, "fmtp:", (size_t) 5)) {
		/* Format parameters:  Not supported */
		/* Note: This is used for codec parameters, like bitrate for
			G722 and video formats for H263 and H264
			See RFC2327 for an example */
		if (debug)
			ast_verbose("Got unsupported a:fmtp in SDP offer \n");
		found = TRUE;
	} else if (!strncasecmp(a, "framerate:", (size_t) 10)) {
		/* Video stuff:  Not supported */
		if (debug)
			ast_verbose("Got unsupported a:framerate in SDP offer \n");
		found = TRUE;
	} else if (!strncasecmp(a, "maxprate:", (size_t) 9)) {
		/* Video stuff:  Not supported */
		if (debug)
			ast_verbose("Got unsupported a:maxprate in SDP offer \n");
		found = TRUE;
	} else if (!strncasecmp(a, "crypto:", (size_t) 7)) {
		/* SRTP stuff, not yet supported */
		if (debug)
			ast_verbose("Got unsupported a:crypto in SDP offer \n");
		found = TRUE;
	}

	return found;
}

static int process_sdp_a_sendonly(char *a, int *sendonly)
{
	int found = FALSE;

	if (!strcasecmp(a, "sendonly")) {
		if (sendonly == -1)
			sendonly = 1;
		found = TRUE;
	} else if (!strcasecmp(a, "inactive")) {
		if (sendonly == -1)
			sendonly = 2;
		found = TRUE;
	}  else if (!strcasecmp(a, "sendrecv")) {
		if (sendonly == -1)
			sendonly = 0;
		found = TRUE;
	}
	return found;
}

static int process_sdp_a_audio(char *a, struct sip_pvt *p, struct ast_rtp *newaudiortp, int *last_rtpmap_codec)
{
	int found = FALSE;
	char* mimeSubtype = ast_strdupa(a); /* ensures we have enough space */

	if (strlen(a) > 5 && !strncasecmp(a, "ptime", 5)) {
		char *tmp = strrchr(a, ':');
		long int framing = 0;
		if (tmp) {
			tmp++;
			framing = strtol(tmp, NULL, 10);
			if (framing == LONG_MIN || framing == LONG_MAX) {
				framing = 0;
				if (option_debug)
					ast_log(LOG_DEBUG, "Can't read framing from SDP: %s\n", a);
			}
		}
		if (framing && p->autoframing) {
			struct ast_codec_pref *pref = ast_rtp_codec_getpref(p->rtp);
			int codec_n;
			int format = 0;
			for (codec_n = 0; codec_n < MAX_RTP_PT; codec_n++) {
				format = ast_rtp_codec_getformat(codec_n);
				if (!format)    /* non-codec or not found */
					continue;
				if (option_debug)
					ast_log(LOG_DEBUG, "Setting framing for %d to %ld\n", format, framing);
				ast_codec_pref_setsize(pref, format, framing);
			}
			ast_rtp_codec_setpref(p->rtp, pref);
		}
		found = TRUE;
	} else if (sscanf(a, "rtpmap: %30u %[^/]/", &codec, mimeSubtype) == 2) {
		/* We have a rtpmap to handle */
		if (last_rtpmap_codec < SDP_MAX_RTPMAP_CODECS) {
			/* Note: should really look at the 'freq' and '#chans' params too */
			if(ast_rtp_set_rtpmap_type(newaudiortp, codec, "audio", mimeSubtype,
				ast_test_flag(&p->flags[0], SIP_G726_NONSTANDARD) ? AST_RTP_OPT_G726_NONSTANDARD : 0) != -1) {
				if (debug)
					ast_verbose("Found audio description format %s for ID %d\n", mimeSubtype, codec);
				last_rtpmap_codec++;
				found = TRUE;
			}
		} else {
			if (debug)
				ast_verbose("Discarded description format %s for ID %d\n", mimeSubtype, codec);
		}

		if (!found) {
			/* Remove this codec since it's an unknown media type for us */
			ast_rtp_unset_m_type(newaudiortp, codec);
			if (debug)
				ast_verbose("Found unknown media description format %s for ID %d\n", mimeSubtype, codec);
		}
	}

	return found;
}

static int process_sdp_a_video(char *a, struct sip_pvt *p, struct ast_rtp *newvideortp, int *last_rtpmap_codec)
{
	int found = FALSE;
	char* mimeSubtype = ast_strdupa(a); /* ensures we have enough space */

	if (sscanf(a, "rtpmap: %30u %[^/]/", &codec, mimeSubtype) == 2) {
		/* We have a rtpmap to handle */
		if (last_rtpmap_codec < SDP_MAX_RTPMAP_CODECS) {
			/* Note: should really look at the 'freq' and '#chans' params too */
			if(p->vrtp && ast_rtp_set_rtpmap_type(newvideortp, codec, "video", mimeSubtype, 0) != -1) {
				if (debug)
					ast_verbose("Found video description format %s for ID %d\n", mimeSubtype, codec);
				last_rtpmap_codec++;
				found = TRUE;
			}
		} else {
			if (debug)
				ast_verbose("Discarded description format %s for ID %d\n", mimeSubtype, codec);
		}

		if (!found) {
			/* Remove this codec since it's an unknown media type for us */
			ast_rtp_unset_m_type(newvideortp, codec);
			if (debug)
				ast_verbose("Found unknown media description format %s for ID %d\n", mimeSubtype, codec);
		}
	}

	return found;
}

static int process_sdp_a_image(char *a, struct sip_pvt *p)
{
	int found = FALSE;
	int peert38capability = 0;
	char s[256];
	int x;

	/* Scan trough the a= lines for T38 attributes and set apropriate fileds */
	if ((sscanf(a, "T38FaxMaxBuffer:%30d", &x) == 1)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "MaxBufferSize:%d\n",x);
	} else if ((sscanf(a, "T38MaxBitRate:%30d", &x) == 1) || (sscanf(a, "T38FaxMaxRate:%30d", &x) == 1)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG,"T38MaxBitRate: %d\n",x);
		switch (x) {
		case 14400:
			peert38capability |= T38FAX_RATE_14400 | T38FAX_RATE_12000 | T38FAX_RATE_9600 | T38FAX_RATE_7200 | T38FAX_RATE_4800 | T38FAX_RATE_2400;
			break;
		case 12000:
			peert38capability |= T38FAX_RATE_12000 | T38FAX_RATE_9600 | T38FAX_RATE_7200 | T38FAX_RATE_4800 | T38FAX_RATE_2400;
			break;
		case 9600:
			peert38capability |= T38FAX_RATE_9600 | T38FAX_RATE_7200 | T38FAX_RATE_4800 | T38FAX_RATE_2400;
			break;
		case 7200:
			peert38capability |= T38FAX_RATE_7200 | T38FAX_RATE_4800 | T38FAX_RATE_2400;
			break;
		case 4800:
			peert38capability |= T38FAX_RATE_4800 | T38FAX_RATE_2400;
			break;
		case 2400:
			peert38capability |= T38FAX_RATE_2400;
			break;
		}
	} else if ((sscanf(a, "T38FaxVersion:%30d", &x) == 1)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "FaxVersion: %d\n",x);
		if (x == 0)
			peert38capability |= T38FAX_VERSION_0;
		else if (x == 1)
			peert38capability |= T38FAX_VERSION_1;
	} else if ((sscanf(a, "T38FaxMaxDatagram:%30d", &x) == 1) || (sscanf(a, "T38MaxDatagram:%30d", &x) == 1)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "FaxMaxDatagram: %d\n",x);
		ast_udptl_set_far_max_datagram(p->udptl, x);
		ast_udptl_set_local_max_datagram(p->udptl, x);
	} else if ((strncmp(a, "T38FaxFillBitRemoval", 20) == 0)) {
		found = TRUE;
		if ((sscanf(a, "T38FaxFillBitRemoval:%30d", &x) == 1)) {
		    if (option_debug > 2)
			ast_log(LOG_DEBUG, "FillBitRemoval: %d\n",x);
		    if (x == 1)
			peert38capability |= T38FAX_FILL_BIT_REMOVAL;
		} else {
		    if (option_debug > 2)
			ast_log(LOG_DEBUG, "FillBitRemoval\n");
		    peert38capability |= T38FAX_FILL_BIT_REMOVAL;
		}
	} else if ((strncmp(a, "T38FaxTranscodingMMR", 20) == 0)) {
		found = TRUE;
		if ((sscanf(a, "T38FaxTranscodingMMR:%30d", &x) == 1)) {
		    if (option_debug > 2)
			ast_log(LOG_DEBUG, "Transcoding MMR: %d\n",x);
		    if (x == 1)
			peert38capability |= T38FAX_TRANSCODING_MMR;
		} else {
		    if (option_debug > 2)
			ast_log(LOG_DEBUG, "Transcoding MMR\n");
		    peert38capability |= T38FAX_TRANSCODING_MMR;
		}
	} else if ((strncmp(a, "T38FaxTranscodingJBIG", 21) == 0)) {
		found = TRUE;
		if ((sscanf(a, "T38FaxTranscodingJBIG:%30d", &x) == 1)) {
		    if (option_debug > 2)
			ast_log(LOG_DEBUG, "Transcoding JBIG: %d\n",x);
		    if (x == 1)
			peert38capability |= T38FAX_TRANSCODING_JBIG;
		} else {
		    if (option_debug > 2)
			ast_log(LOG_DEBUG, "Transcoding JBIG\n");
		    peert38capability |= T38FAX_TRANSCODING_JBIG;
		}
	} else if ((sscanf(a, "T38FaxRateManagement:%255s", s) == 1)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "RateManagement: %s\n", s);
		if (!strcasecmp(s, "localTCF"))
			peert38capability |= T38FAX_RATE_MANAGEMENT_LOCAL_TCF;
		else if (!strcasecmp(s, "transferredTCF"))
			peert38capability |= T38FAX_RATE_MANAGEMENT_TRANSFERED_TCF;
	} else if ((sscanf(a, "T38FaxUdpEC:%255s", s) == 1)) {
		found = TRUE;
		if (option_debug > 2)
			ast_log(LOG_DEBUG, "UDP EC: %s\n", s);
		if (!strcasecmp(s, "t38UDPRedundancy")) {
			peert38capability |= T38FAX_UDP_EC_REDUNDANCY;
			ast_udptl_set_error_correction_scheme(p->udptl, UDPTL_ERROR_CORRECTION_REDUNDANCY);
		} else if (!strcasecmp(s, "t38UDPFEC")) {
			peert38capability |= T38FAX_UDP_EC_FEC;
			ast_udptl_set_error_correction_scheme(p->udptl, UDPTL_ERROR_CORRECTION_FEC);
		} else {
			peert38capability |= T38FAX_UDP_EC_NONE;
			ast_udptl_set_error_correction_scheme(p->udptl, UDPTL_ERROR_CORRECTION_NONE);
		}
	}

	if (found) { /* Some cisco equipment returns nothing beside c= and m= lines in 200 OK T38 SDP */
		p->t38.peercapability = peert38capability;
		p->t38.jointcapability = (peert38capability & 255); /* Put everything beside supported speeds settings */
		peert38capability &= (T38FAX_RATE_14400 | T38FAX_RATE_12000 | T38FAX_RATE_9600 | T38FAX_RATE_7200 | T38FAX_RATE_4800 | T38FAX_RATE_2400);
		p->t38.jointcapability |= (peert38capability & p->t38.capability); /* Put the lower of our's and peer's speed */
	}
	if (debug)
		ast_log(LOG_DEBUG, "Our T38 capability = (%d), peer T38 capability (%d), joint T38 capability (%d)\n",
			p->t38.capability,
			p->t38.peercapability,
			p->t38.jointcapability);

	return found;
}


